﻿#!/usr/bin/perl
use utf8;

# このスクリプトは
#
# Perl正規表現雑技
# http://www.din.or.jp/~ohzaki/regex.htm
#
# で使われているサブルーチン
#
# sub rsrt
# sub substitute
# sub resolv
#
# の３つを改変して正規表現を作る過程を見られるようにしたものです
# 改変したサブルーチン名には "_d" を付けてあります ( デバックのd )
#
# また、処理の過程で用済みになった行列の成分は "-" に置換されます
#
# このスクリプトの 利用・改造は Perl正規表現雑技 と同じく自由です
# 難解な行列処理の理解に繋がれば幸いです
#
#
# 2018/10/11  宮沢 とん子




# デバックモード   @matrix が変更されるたびに print する機能
$debug_mode = 1; # 0 はオフ




$states = 4;
@matrix = (     # E A D C B
              [qw(X X X a b)], # A = Bb + Ca
              [qw(X X X b a)], # D = Ba + Cb
              [qw(X a b X X)], # C = Aa + Db
              [qw(E b a X X)], # B = Ab + Da + e
           );
           #  １行２列の成分は $matrix[0][1]


#$states = 3;
#@matrix = ([qw(E [0369] [258] [147])], # A = A0 + B2 + C1 + e
#           [qw(X [147] [0369] [258])], # B = A1 + B0 + C2
#           [qw(X [258] [147] [0369])], # C = A2 + B1 + C0
#           );



#$states = 9;
#@matrix = ([qw(E [09] 8 7 6 5 4 3 2 1)],
#           [qw(X 1 [09] 8 7 6 5 4 3 2)],
#           [qw(X 2 1 [09] 8 7 6 5 4 3)],
#           [qw(X 3 2 1 [09] 8 7 6 5 4)],
#           [qw(X 4 3 2 1 [09] 8 7 6 5)],
#           [qw(X 5 4 3 2 1 [09] 8 7 6)],
#           [qw(X 6 5 4 3 2 1 [09] 8 7)],
#           [qw(X 7 6 5 4 3 2 1 [09] 8)],
#           [qw(X 8 7 6 5 4 3 2 1 [09])],
#           );



###  ある行を他の行に代入する
sub substitute_d {

   ## $states = 4;  のとき、
   ## 代入元の行 $y = 3 2 1   の順に入る
   ## 処理する行 $i = 0 1 2   の順に入る
   ## 処理する列 $j = 0 1 2 3 の順に入る

   my $y = shift;

   # 各行に代入
   foreach my $i (0 .. $y - 1 ){

      # 代入先の行に代入元の行の成分が無いとき
      if ( $matrix[$i][$y+1] eq 'X' ){
         $matrix[$i][$y+1] = '-';  #  X を - に書き換え
         dbug1( $debug_mode );
         next;  # 以降の処理をスキップして $i を1つ進める
      }

      # $i 行の各成分に代入する
      foreach my $j (0 .. $y ){

         if ( $matrix[$y][$j] ne 'X' ){   # 代入元の成分があれば
            # 代入先の成分の有無で代入するものを変え、代入する
            if ( $matrix[$i][$j] eq 'X' ){
                $matrix[$i][$j] = "$matrix[$y][$j]$matrix[$i][$y+1]";
                dbug1( $debug_mode );
             } else {
                $matrix[$i][$j] = "($matrix[$i][$j]|$matrix[$y][$j]$matrix[$i][$y+1])";
                dbug1( $debug_mode );
            }  # if
         }  # if

      }  # foreach $j

      # $i 行への代入が終わったら $i 行にある代入元の成分を消す
      $matrix[$i][$y+1] = '-';
      dbug1( $debug_mode );

   }  # foreach $i

   # 代入が終わったら代入元の行を '-' にする
   foreach my $j (0 .. $y ){
      $matrix[$y][$j] = '-';
   }  # foreach
   dbug1( $debug_mode );

}  # sub substitute_d




###  r = s + rt → r = st*
sub rsrt_d {
  my $j = shift;  # $state = 4;  のとき $j = 3,2,1,0 の4回呼び出される

  #  例えば  B = Ag + Ch + Bq   のように自分を含む成分があれば
  if ($matrix[$j][$j + 1] ne 'X')
  {
     #  B = Agq* + Chq*  に変形させる
     foreach my $i (0 .. $j)
     {
         if ($matrix[$j][$i] ne 'X')
         {
             $matrix[$j][$i] .= "$matrix[$j][$j + 1]*";
             dbug1( $debug_mode );

         }  # if
     }  # foreach
  }  # if
  $matrix[$j][$j + 1] = '-';  # Bq は用済みなので消す
  dbug1( $debug_mode );
}  # sub rsrt



###  $matrix[0][0]  に正規表現を作成する
sub resolv_d {
   dbug1( $debug_mode );  # 初期状態の表示

   for (my $j = $states - 1; $j > 0; $j--) {
      rsrt_d($j);
      substitute_d($j);
   }
   rsrt_d(0);  # $j = 0 のときだけ substitute を実行させない
               # 代入先となる行が存在しないため

   $matrix[0][0] =~ s/E//g;
   dbug1( $debug_mode );

   $matrix[0][0] = '^' . $matrix[0][0] . '$';
   dbug1( $debug_mode );

   return $matrix[0][0];  # 戻り値
}  # resolv_d




###  デバック  @matrix を print する
sub dbug1 {
    # デバックモードがオフなら止める
    if  ( shift == 0 ){ return; }

    foreach my $i ( 0 .. $states ){
       foreach my $j ( 0 .. $states + 1 ){
          print " $matrix[$i][$j]";  # 各成分の前に半角スペを入れる
       }  # foreach
       print "\n";  # 行の最後に改行を入れる
    }  # foreach
    print "\n";  # 行列の後ろに空行を入れる
}  # sub dbug1





### 余計なカッコを削除する ( 改変なし )
sub simplify {
  my $expr = shift;
  $openclose = qr/[^()]*(?:\((??{$openclose})\)[^()]*)*/;
  $no_or = qr/[^(|)]*(?:\((??{$openclose})\)[^(|)]*)*/;
  $term = qr/(?:(?:\w+|\((??{$term})(?:\|(??{$term}))*\))\*?)/;
  use re 'eval';
  1 while (0 or
           $expr =~ s/\((\w)\)/$1/g or # (a)→a
           $expr =~ s/\(($no_or)\)(?![*?])/$1/g or # (expr)→expr
           $expr =~ s/^\(($openclose)\)$/$1/g or # ^(expr)$→expr
           $expr =~ s/\((\($openclose\))\)/$1/g or # ((expr))→(expr)
           $expr =~ s/((?:^|\()(?:$term+\|)*)
           \(
           ($term+(?:\|$term+)+)
           \)
           ((?:\|$term+)*(?:\)|$))/$1$2$3/gx or
	    # (expr1|(expr2|expr3))→(expr1|expr2|expr3)

           $expr =~ s/\((\w)[*?]\)\*/$1*/g or # (a[*?])*→a*
           $expr =~ s/\((\w)\*\)[*?]/$1*/g or # (a*)[*?]→a*
           $expr =~ s/(\w)[*?]\1\*/$1*/g or # a[*?]a*→a*
           $expr =~ s/(\w)\*\1[*?]/$1*/g or # a*a[*?]→a*
           
           $expr =~ s/\(($openclose)\)[*?]\(\1\)\*/($1)*/g or
	    # (expr)[*?](expr)*→(expr)*
           $expr =~ s/\(($openclose)\)\*\(\1\)[*?]/($1)*/g or
	    # (expr)*(expr)[*?]→(expr)*
           $expr =~ s/\((\($openclose\))[*?]\)\*/$1*/g or
	    # ((expr)[*?])*→(expr)*
           $expr =~ s/\((\($openclose\)\*)\)[*?]/$1/g or
	    # ((expr)*)[*?]→(abc)*
           $expr =~ s/(^|[(|])($no_or)((?:\|$no_or)*)\|\2($|[|)])/$1$2$3$4/
	    or # expr1|expr2|expr1→expr1|expr2
           $expr =~ s/(^|[(|])($no_or)\|($no_or)($|[|)])
                      (?(?{(length $2 > length $3 or
			   (length $2 == length $3 and ($2 cmp $3) > 0))
                          })|(?!))/$1$3|$2$4/x
	    or # expr1|expr2→sort(expr1|expr2)
           0);
  no re 'eval';
  $expr;
}


$result = resolv_d();
$result = simplify($result);
print $result;







# Perl正規表現雑技
# http://www.din.or.jp/~ohzaki/regex.htm


